﻿DeclareModule DS18B20

  ;-DeclareModule

  Enumeration
    #DS_Err_OK   ; Ошибок нет.
    #DS_Err_Port ; Нет открытого порта (для OpenPort() - ошибка открытия порта).
    #DS_Err_NoSensor ; Не найден датчик.
    #DS_Err_IO   ; Ошибка при обмене информацияей с датчиком.
    #DS_Err_BaudRate ; Ошибка при изменении скорости порта.
  EndEnumeration
  
  Structure DS_DataByte ; Копия содержимого ОЗУ датчика
    L_Termo.a       ; Младший байт температуры
    H_Termo.a       ; Старший байт температуры
    H_User.a        ; Вехрний предел температуры
    L_User.a        ; Нижний предел температуры
    Config.a        ; Байт конфигурации
    x1.a            ; Резерв (в датчике не реализовано)
    x2.a            ; Резерв (в датчике не реализовано)
    x3.a            ; Резерв (в датчике не реализовано)
    CRC.a           ; Контрольная сумма
  EndStructure
  
  #PortOnly = -10
  
  Declare GetLastErr()
  Declare.s GetLastErr_String(Err=-1)
  Declare ClosePort(Port=#PortOnly)
  Declare OpenPort(ComPort.s, Port=#PortOnly)
  Declare ResetDS(Port=#PortOnly)
  Declare.a ByteRW(Byte.a=$FF, Port=#PortOnly)
  Declare.a GetCRC(*Buff, Count)
  Declare GetThermo(*Info.DS_DataByte, Port=#PortOnly)
  
EndDeclareModule

Module DS18B20

EnableExplicit

;- Private

Structure CRC_Buff
  Byte.a[0]
EndStructure

CompilerIf #PB_Compiler_OS = #PB_OS_Linux
  #NCCS = 32
  #TCSANOW = 0
  
  Structure termios Align #PB_Structure_AlignC
    c_iflag.l     ; .l because .i changes size with the x64 version
    c_oflag.l
    c_cflag.l
    c_lflag.l
    c_line.a
    c_cc.a[#NCCS]
    c_ispeed.l
    c_ospeed.l
  EndStructure
CompilerEndIf

Threaded tLastErr=0
Global gCurrent_ComPort_ID=-1

Macro TestPortID(pID)
  If pID=#PortOnly
    pID=gCurrent_ComPort_ID
  EndIf
EndMacro

Macro Odd(xx)
  ((xx) & 1)
EndMacro

Procedure ClearInBuf(Port) ; Очистка приёмного буфера.
  Protected Count
  
  Count = AvailableSerialPortInput(Port)
  If Count>0
    Protected Dim Buff.a(Count+2)
    ReadSerialPortData(Port, @Buff(), Count)
  EndIf
EndProcedure

Procedure SetBaudRate(Port, Rate)
  Protected Result=#False, PortID

  If IsSerialPort(Port)
    
    PortID = SerialPortID(Port)
    
    CompilerIf #PB_Compiler_OS = #PB_OS_Windows
      
      Protected dcb.DCB
      If PortID
        If GetCommState_(PortID, @dcb)  
          dcb\BaudRate  = Rate ; Требуемая скорость.
          dcb\fbits | %101000000010001
          If SetCommState_(PortID, @dcb)
            Result = #True
          EndIf
        EndIf
      EndIf
      
    CompilerElse
      
      Protected options.termios, Sp
      
      If Rate=9600 Or Rate=115200
        
        Select Rate
          Case 9600   : Sp = 13
          Case 115200 : Sp = 4098
        EndSelect
        
        If tcgetattr_(PortID, @options) = 0   ; читает параметры порта
          If cfsetispeed_(@options, Sp) = 0   ; установка скорости порта
            If cfsetospeed_(@options, Sp) = 0 ; установка скорости порта
              If tcsetattr_(PortID, #TCSANOW, @options) = 0
                Result=#True
              EndIf
            EndIf
          EndIf
        EndIf
        
      EndIf
      
    CompilerEndIf
    
    If Result=#False : tLastErr=#DS_Err_BaudRate : EndIf
  EndIf
    
  ProcedureReturn Result
EndProcedure


;-Public

Procedure GetLastErr()
  ProcedureReturn tLastErr
EndProcedure

Procedure.s GetLastErr_String(Err=-1)
  Protected String.s
  
  If Err=-1
    Err=tLastErr
  EndIf
  Select Err
    Case #DS_Err_OK : String = "Нет ошибок"
    Case #DS_Err_Port : String = "Не удалось открыть порт"
    Case #DS_Err_NoSensor : String = "Не найден датчик температуры"
    Case #DS_Err_IO : String = "Ошибка обмена данными с датчиком"
    Case #DS_Err_BaudRate : String = "Ошибка изменения скорости порта"
  EndSelect
    
  ProcedureReturn String
EndProcedure

Procedure ClosePort(Port=#PortOnly)
  Protected Result = #False
  
  TestPortID(Port)
  
  If IsSerialPort(Port)
    SetSerialPortStatus(Port, #PB_SerialPort_DTR, 0)
    SetSerialPortStatus(Port, #PB_SerialPort_RTS, 0)
    CloseSerialPort(Port)
    Result = #True
    tLastErr = #DS_Err_OK
  Else
    tLastErr = #DS_Err_Port
  EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure OpenPort(ComPort.s, Port=#PortOnly)
  Protected ID, PortID
  
  If Port>=0
    ClosePort(Port)
    PortID=Port
  Else
    PortID=#PB_Any
  EndIf
  
  ID = OpenSerialPort(PortID, ComPort, 115200, #PB_SerialPort_NoParity,
                      8, 1, #PB_SerialPort_RtsHandshake, 24, 24)
  If ID
    If Port=#PortOnly
      gCurrent_ComPort_ID = ID
    ElseIf Port>=0
      ID=Port
    EndIf
    SetSerialPortStatus(ID, #PB_SerialPort_DTR, 1)
    SetSerialPortStatus(ID, #PB_SerialPort_RTS, 1)
    tLastErr = #DS_Err_OK
  Else
    tLastErr=#DS_Err_Port
  EndIf
  
  ProcedureReturn ID
EndProcedure

Procedure ResetDS(Port=#PortOnly) ; Подаём команду "Сброс" датчику температуры.
  Protected x, Result=#False
  
  TestPortID(Port)
  
  tLastErr=#DS_Err_OK
  
  If IsSerialPort(Port)
    If SetBaudRate(Port, 9600)
      
      ClearInBuf(Port)
      
      x=$F0
      WriteSerialPortData(Port, @x, 1)
      x = ElapsedMilliseconds()+100
      Repeat
        Delay(2)
      Until AvailableSerialPortInput(Port)>0 Or ElapsedMilliseconds()>x
      
      If AvailableSerialPortInput(Port)=1
        x=0
        If ReadSerialPortData(Port, @x, 1) = 1
          Result=Bool(x<>$F0)
          If Result=#False : tLastErr=#DS_Err_NoSensor : EndIf
        Else
          tLastErr=#DS_Err_NoSensor
        EndIf
      Else
        tLastErr=#DS_Err_NoSensor
      EndIf
      
      SetBaudRate(Port, 115200)
      
    EndIf
  Else
    tLastErr=#DS_Err_Port
  EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure.a ByteRW(Byte.a=$FF, Port=#PortOnly)
  Protected Result.a, CRC.a, x.a, i, Time
  
 
  TestPortID(Port)
  
  tLastErr=#DS_Err_OK
  Result=$FF
  
  If IsSerialPort(Port)
    ClearInBuf(Port)
    
    Time=ElapsedMilliseconds()+100
    
    For i= 1 To 8
      
      x=Odd(Byte)*$FF
      WriteSerialPortData(Port, @x, 1)
      Byte>>1
      
      Repeat : Until AvailableSerialPortInput(Port)>0 Or ElapsedMilliseconds()>Time
      
      If AvailableSerialPortInput(Port)=1
        x=0
        If ReadSerialPortData(Port, @x, 1) = 1
          Result>>1
          If Odd(x) : Result | $80 : EndIf
        Else
          Result = $FF
          tLastErr=#DS_Err_IO
          Break
        EndIf
      Else
        Result = $FF
        tLastErr=#DS_Err_IO
        Break
      EndIf
      
    Next i
    
  Else
    tLastErr=#DS_Err_Port
  EndIf

  ProcedureReturn Result
EndProcedure

Procedure.a GetCRC(*Buff.CRC_Buff, Count)
  Protected CRC.a, Byte.a, i, x
  
  Count-1
  For i=0 To Count
    Byte=*Buff\Byte[i]
    For x=1 To 8
      If (Byte ! CRC) & 1
        CRC = ((CRC!$18)>>1)|$80
      Else
        CRC >> 1
      EndIf
      Byte>>1
    Next x
  Next i
  
  ProcedureReturn CRC
EndProcedure

Procedure GetThermo(*Info.CRC_Buff, Port=#PortOnly)
  Protected i, Result=#False
  
  tLastErr=#DS_Err_OK
  
  TestPortID(Port)
  If IsSerialPort(Port)
    ClearInBuf(Port)
    
    If ResetDS(Port)
      ByteRW($CC, Port)
      ByteRW($44, Port)
      Delay(800)
      If ResetDS(Port)
        ByteRW($CC, Port)
        ByteRW($BE, Port)
        
        For i=0 To 8
          *Info\Byte[i] = ByteRW($FF, Port)
        Next
        
        Result=#True
        
      EndIf
    EndIf
    
  Else
    tLastErr=#DS_Err_Port
  EndIf
    
  ProcedureReturn Result
EndProcedure

EndModule
; IDE Options = PureBasic 5.21 LTS (Windows - x86)
; CursorPosition = 117
; FirstLine = 93
; Folding = ---
; EnableUnicode
; EnableXP